UNIT MATRIX;

(*
       Ŀ
            This is a part of the original source text for the    
              program DisCo, version 1.1 - December 10, 1993      
                                                                  
                                                
                             MATRIX.PAS 1.1                       
                                                
                                                                  
         If you modify this unit, please                          
                                                                  
            (1) Make sure that you have a copy of the original    
                file.                                             
                                                                  
            (2) Delete this comment (25 lines).                   
                                                                  
         Original size of file:  17488 bytes.                     
       
             Copyright       Bent Nielsen
                             Institute of Mathematical Statistics
                             University of Copenhagen
                             Universitetsparken 5
                             2100 Copenhagen O
                             DK - Denmark
*)

{---- Matrix Calculation ------------------}

{---- Interface ---------------------------}

{$D+}

INTERFACE

CONST    Tmax = 800;
         pmax = 20;
         delta= 1E-10;   { definition of computer zero }


TYPE
         T_Vector    = Array[-1..Tmax] Of Single;
         pxT_matrix  = Array[1..pmax]  Of ^T_Vector;
         pxp_matrix  = Array[1..pmax]  Of Array[1..pmax] of Double;

PROCEDURE CholeskyDecomposition(var L :pxp_matrix; var Dim :Byte;
                                var OK :Boolean);
  { 4.10.1993
    Input:  L: the lower triangular part of an positive definite matrix A, say.
    Output: L: the Cholesky decomposition of A, an lower triangular matrix.
            OK: False if matrix not is positive definite. Note that symmetri
                is assumed and not checked.
    }

PROCEDURE GetMatrix(var Matrix :pxT_matrix; var Dim :Byte);

PROCEDURE DeleteMatrix(var Matrix :pxT_matrix; var Dim :Byte);

PROCEDURE CalculateInvL11S10(var X, DeltaX :pxT_Matrix; var T :Word;
                             var DimX, DimDeltaX :Byte; var Res :pxp_matrix);

PROCEDURE MultipelRegression(var X, Reg, Res : pxT_matrix; var T: word;
                             var DimX, DimReg :Byte;
                             var InvRegReg : pxp_matrix);

PROCEDURE TraceOfSquareToMatrix(var D :pxp_matrix; var m,n :Byte;
                                  var Trace :Single);

PROCEDURE SquareToMatrix(var X:pxT_matrix; var Res:pxp_matrix;
                         var T:word; var p:Byte);

PROCEDURE InversePosDefMatrix(var A :pxp_matrix; var Dim :Byte);
  { This procedure reads the lower triangular part of an positive definite
    matrix and inverses the matrix.  The result is not only the lower
    triangular part but the hole matrix. }

{---- Implementation ----------------------}

IMPLEMENTATION

USES Crt;

(* - - - - Run Time Errors *)

PROCEDURE Error(Code :Byte);
VAR      ErrorText: String;
Begin
  Case Code Of
   20:  ErrorText:= 'RUN TIME ERROR 20: Inversion of singular matrix';
                                   { Obtained in the PROCEDURE CholeskyDecomp }
   101: ErrorText:= 'RUN TIME ERROR 101: in eigenvalue procedure';
                   { Obtained in the PROCEDURE EigenValuesOfTriDiagonalMatrix }
  End;
  If Code > 0 Then
    Begin
      Writeln;
      Writeln(ErrorText);
      HALT;
    End;
End;




(* - - - - initialization of matrices
PROCEDURE GetMatrix(var Matrix :pxT_matrix; var Dim :Byte);

PROCEDURE DeleteMatrix(var Matrix :pxT_matrix; var Dim :Byte);
- - - - *)

PROCEDURE GetMatrix(var Matrix :pxT_matrix; var Dim :Byte);
Var      i:Byte;
BEGIN { proc GetMatrix }
  For i:= 1 To Dim Do
    New(Matrix[i]);
END;  { proc GetMatrix }

PROCEDURE DeleteMatrix(var Matrix :pxT_matrix; var Dim :Byte);
Var      i:Byte;
BEGIN { proc DeleteMatrix }
  For i:= 1 To Dim Do
    Dispose(Matrix[i]);
END;  { proc DeleteMatrix }


(* - - - - multiplication and subtraction of matrices - - - - -

PROCEDURE MultiplicateLargeMatrices(var X,Y:Txp_matrix; var Res:pxp_matrix;
                                    var k:Byte; var m:Word; var n:Byte);
  { Input:  X^t is a kxm matrix and Y is mxn matrix.
    OutPut: Res = X^t*Y is a kxn matrix. }
PROCEDURE MultiplicateLargeAndSmallMatrices(var X:Txp_matrix; var Y:pxp_matrix;
                                    var Res:Txp_matrix;
                                    var k:Word; var m,n:Byte);
  { Input:  X is a kxm matrix and Y is mxn matrix.
    OutPut: Res = X*Y is a kxn matrix. }
PROCEDURE MultiplicateSmallMatrices(var X,Y,Res:pxp_matrix; var k,m,n:Byte);
  { Input:  X is a kxm matrix and Y is mxn matrix.
    OutPut: Res = X*Y is a kxn matrix. }
PROCEDURE MultiplicateTriangularAndSmallMatrices(var X,Y,Res:pxp_matrix;
                                                 var k,n:byte);
  { Input:  X is a lower triangular kxk matrix and Y is kxn matrix.
    OutPut: Res = X*Y is a kxn matrix. }
PROCEDURE SquareLowerTriangularMatrix(var A :pxp_matrix; var Dim :Byte);
  { Input:  A: an lower triangular matrix L, say.
    OutPut: A: L^t * L.
            The result is not only the lower triangular part but
            the hole matrix. }
PROCEDURE SquareToMatrix(var X:Txp_matrix; var Res:pxp_matrix;
                         var T:word; var p:Byte);
  { Input:  X is a Txp matrix
    OutPut: Res = X^t * X is a postive definite pxp matrix.
            Remark the lower triangular part is stored. }
PROCEDURE SquareToSmallMatrix(var X:pxp_matrix; var Res:pxp_matrix;
                         var m,n:Byte);
  { Input:  X is a mxn matrix
    OutPut: Res = X^t * X is a nxn matrix.
            Remark only the lower triangular part is stored. }
PROCEDURE SubtractMatrix(var X,Y,Res: Txp_matrix; var m:Word; var n:Byte);
  { Input:  X, Y are mxn matrices.
    OutPut: Res = X - Y is a mxn matrix }
PROCEDURE SubtractMatrix(var X,Y,Res: Txp_matrix; var m:Word; var n:Byte);
  { Input:  X, Y are mxn matrices.
    OutPut: Res = X - Y is a mxn matrix }

- - - - - - - - - *)

PROCEDURE MultiplicateLargeMatrices(var X,Y:pxT_matrix; var Res:pxp_matrix;
                                    var pX:Byte; var T:Word; var pY:Byte);
  { Input:  X is a pX x T matrix and Y^t is a T x pY matrix.
    OutPut: Res = X*Y^t is a pX x pY matrix. }
  var   Row,Col:byte;
        i:word;
        Dum:Double;
BEGIN {  MultiplicateLargeMatrices }
  For Row:= 1 To pX Do
    For Col:= 1 to pY Do
      Begin
        Dum:= 0;
        For i:=1 To T Do
          Dum:= Dum + X[Row]^[i] * Y[Col]^[i];
        Res[Row,Col]:= Dum;
      End;
END;  {  MultiplicateLargeMatrices }

PROCEDURE MultiplicateSmallAndLargeMatrices(var X:pxp_matrix; var Y:pxT_matrix;
                                    var Res:pxT_matrix;
                                    var pX,pY:Byte; var T:Word);
  { Input:  X is a pX x pY matrix and Y is a pY x T matrix.
    OutPut: Res   = X * Y is a pX x T matrix . }
  var   Col:word;
        Row,i:byte;
        Dum:Double;
BEGIN {  MultiplicateLargeMatrices }
  For Row:= 1 To pX Do
    For Col:= 1 to T Do
      Begin
        Dum:= 0;
        For i:=1 To pY Do
          Dum:= Dum + X[Row][i] * Y[i]^[Col];
        Res[Row]^[Col]:= Dum;
      End;
END;  {  MultiplicateLargeMatrices }

PROCEDURE MultiplicateSmallMatrices(var X,Y,Res:pxp_matrix; var k,m,n:Byte);
  { Input:  X is a k x m matrix and Y is m x n matrix.
    OutPut: Res = X*Y is a kxn matrix. }
  var   Row,Col,i:byte;
        Dum:Double;
BEGIN {  MultiplicateSmallMatrices }
  For Row:= 1 To k Do
    For Col:= 1 to n Do
      Begin
        Dum:= 0;
        For i:=1 To m Do
          Dum:= Dum + X[Row,i] * Y[i,Col];
        Res[Row,Col]:= Dum;
      End;
END;  {  MultiplicateSmallMatrices }

PROCEDURE MultiplicateTriangularAndSmallMatrices(var X,Y,Res:pxp_matrix;
                                                 var k,n:byte);
  { Input:  X is a lower triangular kxk matrix and Y is kxn matrix.
    OutPut: Res = X*Y is a kxn matrix. }
  var   Row,Col,i:byte;
        Dum:Double;
BEGIN { proc MultiplicateTriangularAndSmallMatrices }
  For Row:= 1 To k Do
    For Col:= 1 to n Do
      Begin
        Dum:= 0;
        For i:=1 To Row Do
          Dum:= Dum + X[Row,i] * Y[i,Col];
        Res[Row,Col]:= Dum;
      End;
END;  { proc MultiplicateTriangularAndSmallMatrices }

PROCEDURE SquareLowerTriangularMatrix(var A :pxp_matrix; var Dim :Byte);
  { Input:  A: an lower triangular matrix L, say.
    OutPut: A: L^t * L.
            The result is not only the lower triangular part but
            the hole matrix. }
  Var   Row,Col,i:Byte;
        Dum:Double;
BEGIN { SquareLowerTriangularMatrix }
  For Row:=1 To Dim Do
    For Col:=1 To Row Do
      Begin
        Dum:= 0;
        For i:= Row To Dim Do
          Dum:= Dum + A[i,Col] * A[i,Row];
        A[Row,Col]:= Dum;
        If Row > Col Then
          A[Col,Row]:= Dum;
      End;
END;  { SquareLowerTriangularMatrix }

PROCEDURE SquareToMatrix(var X :pxT_matrix; var Res:pxp_matrix;
                         var T:word; var p:Byte);
  { Input:  X is a pxT matrix
    OutPut: Res = X * X^t is a positive definite pxp matrix.
            Remark only the lower triangular part is stored. }
  var   Row,Col:byte;
        i:word;
        Dum:Double;
BEGIN { proc SquareToMatrix }
  For Row:=1 To p Do
    For Col:= 1 to Row Do
      Begin
        Dum:= 0;
        For i:=1 To T Do
          Dum:= Dum + X[Row]^[i] * X[Col]^[i];
        Res[Row,Col]:= Dum;
      End;
END;  { proc SquareToMatrix }

PROCEDURE SquareToSmallMatrix(var X:pxp_matrix; var Res:pxp_matrix;
                         var m,n:Byte);
  { Input:  X is a mxn matrix
    OutPut: Res = X^t * X is a nxn matrix.
            Remark only the lower triangular part is stored. }
  var   Row,Col,i:byte;
        Dum:Double;
BEGIN { proc SquareToSmallMatrix }
  For Row:= 1 To n Do
    For Col:= 1 to Row Do
      Begin
        Dum:= 0;
        For i:=1 To m Do
          Dum:= Dum + X[i,Row] * X[i,Col];
        Res[Row,Col]:= Dum;
      End;
END;  { proc SquareToSmallMatrix }

PROCEDURE SubtractMatrix(var X,Y,Res: pxT_matrix; var T:Word; var p:Byte);
  { Input:  X, Y are pxT matrices.
    OutPut: Res = X - Y is a pxT matrix }
  Var       Row:Byte;
            Col:Word;
BEGIN { SubtractMatrix }
  For Row:= 1 To p Do
    For Col:= 1 To T Do
      Res[Row]^[Col]:= X[Row]^[Col] - Y[Row]^[Col];
END;  { SubtractMatrix }

PROCEDURE SubtractSmallMatrix(var X,Y,Res: pxp_matrix; var m,n:Byte);
  { Input:  X, Y are mxn matrices.
    OutPut: Res = X - Y is a mxn matrix }
  Var       Row:word;
            Col:Byte;
BEGIN { SubtractSmallMatrix }
  For Row:= 1 To m Do
    For Col:= 1 To n Do
      Res[Row,Col]:= X[Row,Col] - Y[Row,Col];
END;  { SubtractSmallMatrix }

PROCEDURE TransposeMatrix(var A: pxp_matrix; var m,n:Byte);
  { 8-7-93
    InPut:  A is a mxn matrix.
    OutPut: A is the transposed nxm matrix
  }
VAR      Row,Col,Max: Byte;
PROCEDURE SwapVar(var x,y: Double);
VAR      Dum: Double;
BEGIN { sub-PROCEDURE SwapVar }
  Dum:= x;
  x:= y;
  y:= Dum;
END;  { sub-PROCEDURE SwapVar }
BEGIN { PROCEDURE TransposeMatrix }
  If m>n Then Max:= m
         Else Max:= n;
  For Col:= 2 To Max Do
    For Row:= 1 To (Col-1) Do
      SwapVar(A[Row,Col],A[Col,Row]);
END;  { PROCEDURE TransposeMatrix }

(* - - more complicated rutines with matrices - - -
PROCEDURE CalculateInvL11S10(var X, DeltaX :Txp_Matrix; var T :Word;
                             var DimX, DimDeltaX :Byte; var Res :pxp_matrix);
PROCEDURE CholeskyDecomposition(var L :pxp_matrix; var Dim :Byte);
  { Input:  L: the lower triangular part of an positive definite matrix A, say.
    Output: L: the Cholesky decomposition of A, an lower triangular matrix. }
PROCEDURE DeterminantOfPosDefMatrix(var X:pxp_matrix; var dim:byte;
                                    var determinant:Single);
PROCEDURE InverseLowerTriangularMatrix(var L :pxp_matrix; var Dim :Byte);
  { Input:  L: an lower triangular matrix.
    Output: L: the inverse of L. }
PROCEDURE InversePosDefMatrix(var A :pxp_matrix; var Dim :Byte);
  { This procedure reads the lower triangular part of an positive definite
    matrix and inverses the matrix.  The result is not only the lower
    triangular part but the hole matrix. }
PROCEDURE MultipelRegression(var X, Reg, Res : Txp_matrix; var T: word;
                             var DimX, DimReg :Byte);
PROCEDURE TraceOfSquareToMatrix(var D :pxp_matrix; var m,n :Byte;
                                  var Trace :Single);
  { Input:  D is a m x n matrix.
            Now let A be a n x n matrix.  Imagine A = D^t * D.
    OutPut: The trace of the matrix A. }
*)



PROCEDURE CholeskyDecomposition(var L :pxp_matrix; var Dim :Byte;
                                var OK :Boolean);
  { 4.10.1993
    Input:  L: the lower triangular part of an positive definite matrix A, say.
    Output: L: the Cholesky decomposition of A, an lower triangular matrix.
            OK: False if matrix not is positive definite. Note that
                symmetricity is assumed and not checked.
    }
  Var   Row,Col,i:Byte;
BEGIN { CholeskyDecomposition }
  OK:= True;
  For i:= 1 To Dim Do
    Begin
      If L[i,i] <= 0 Then Begin OK:= False; EXIT End;
      L[i,i]:= sqrt(L[i,i]);
      For Row:= (i+1) To Dim Do
        Begin
          L[Row,i]:= L[Row,i]/L[i,i];
          For Col:= (i+1) To Row Do
            L[Row,Col]:= L[Row,Col] - L[Row,i] * L[Col,i];
        End;
    End;
END;  { CholeskyDecomposition }

PROCEDURE DeterminantOfPosDefMatrix(var X:pxp_matrix; var dim:byte;
                                    var determinant:Single);
  Var   Row:Byte;
        OK :Boolean;
BEGIN { proc DeterminantOfPosDefMatrix }
  CholeskyDecomposition(X,Dim,OK);
  If OK=False Then Error(20);
  determinant:=1;
  For Row:= 1 to Dim Do
    determinant:= determinant * sqr(X[Row,Row]);
END;  { proc DeterminantOfPosDefMatrix }

PROCEDURE InverseLowerTriangularMatrix(var L :pxp_matrix; var Dim :Byte);
  { Input:  L: an lower triangular matrix.
    Output: L: the inverse of L. }
  Var   Row,Col,i:Byte;
BEGIN { InverseLowerTriangularMatrix }
  For Row:= 1 To Dim Do
    Begin
      For Col:= 1 To (Row-1) Do
        L[Row,Col]:= L[Row,Col]/L[Row,Row];
      L[Row,Row]:= 1/L[Row,Row];
    End;
  For Row:= Dim DownTo 2 Do
    For Col:= (Row-1) DownTo 1 Do
      Begin
        For i:=1 To (Col-1) Do
          L[Row,i]:= L[Row,i] - L[Row,Col] * L[Col,i];
        L[Row,Col]:= 0 - L[Row,Col] * L[Col,Col];
      End;
END;  { InverseLowerTriangularMatrix }

PROCEDURE CalculateInvL11S10(var X, DeltaX :pxT_Matrix; var T :Word;
                             var DimX, DimDeltaX :Byte; var Res :pxp_matrix);
VAR      S10, S11 :pxp_matrix;
         OK :Boolean;
BEGIN { proc CalculateInvL11S10 }
  SquareToMatrix(X,S11,T,DimX);
  MultiplicateLargeMatrices(X,DeltaX,S10,DimX,T,DimDeltaX);
  CholeskyDecomposition(S11,DimX,OK);
  If OK= False Then Error(20);
  InverseLowerTriangularMatrix(S11,DimX);
  MultiplicateTriangularAndSmallMatrices(S11,S10,Res,DimX,DimDeltaX);
END;  { proc CalculateInvL11S10 }

PROCEDURE InversePosDefMatrix(var A :pxp_matrix; var Dim :Byte);
  { This procedure reads the lower triangular part of an positive definite
    matrix and inverses the matrix.  The result is not only the lower
    triangular part but the hole matrix. }
  VAR      OK : Boolean;
BEGIN { InversePosDefMatrix }
  CholeskyDecomposition(A,Dim,OK);
  IF OK=False Then Error(20);
  InverseLowerTriangularMatrix(A,Dim);
  SquareLowerTriangularMatrix(A,Dim);
END;  { InversePosDefMatrix }

PROCEDURE MultipelRegression(var X, Reg, Res : pxT_matrix; var T: word;
                             var DimX, DimReg :Byte;
                             var InvRegReg : pxp_matrix);
  Var
         XReg, XRegInvRegReg : pxp_matrix;
         Projection   : pxT_matrix;
BEGIN { MultipelRegression }
  MultiplicateLargeMatrices(X,Reg,XReg,DimX,T,DimReg);
  MultiplicateSmallMatrices(XReg,InvRegReg,XRegInvRegReg,DimX,DimReg,DimReg);
  GetMatrix(Projection,DimX);
    MultiplicateSmallAndLargeMatrices(XRegInvRegReg,Reg,Projection,DimX,DimReg,T);
    SubtractMatrix(X,Projection,Res,T,DimX);
  DeleteMatrix(Projection,DimX);
END;  { MultipelRegression }

PROCEDURE TraceOfSquareToMatrix(var D :pxp_matrix; var m,n :Byte;
                                  var Trace :Single);
  { Input:  D is a m x n matrix.
            Now let A be a n x n matrix.  Imagine A = D^t * D.
    OutPut: The trace of the matrix A. }
VAR      Col,Row:Byte;
BEGIN { proc TraceOfSquareToMatrix }
  Trace:=0;
  For Row:= 1 To m Do
    For Col:= 1 To n Do
      Trace:= Trace + Sqr(D[Row,Col]);
END;  { proc TraceOfSquareToMatrix }

END.