UNIT SIM_UNIT;

(*
         Ŀ
              This is a part of the original source text for the    
                  program DisCo, version 1.4 - 18 April 1997        
                                                                    
                                                  
                              SIM_UNIT.PAS 1.4                      
                                                  
                                                                    
           If you modify this unit, please                          
                                                                    
              (1) Make sure that you have a copy of the original    
                  file.                                             
                                                                    
              (2) Delete this comment (25 lines).                   
                                                                    
           Original size of file:  8029 bytes.                      
         
             Copyright       Bent Nielsen
                             Institute of Mathematical Statistics
                             University of Copenhagen
                             Universitetsparken 5
                             2100 Copenhagen O
                             DK - Denmark
*)

{---- Interface ---------------------------}

INTERFACE

{$D+}

PROCEDURE Simulation;

{---- Implementation ----------------------}

IMPLEMENTATION

USES     Sim_IO,Matrix,Crt;

VAR
         Reg,TrendStat :pxT_matrix;
         RegReg :pxp_matrix;
(*16/5         RegSum : T_Vector;          *)
         EraseMe :Text;
         DeltaX, X_ :pxT_matrix;


{ Here comes the procedures
                      CreateDeltaX  and CreateX_
}

PROCEDURE CreateDeltaX;
  { This procedure simulates a DimDeltaX x T matrix of i.i.d. Gaussian r.v. }
VAR      Row :Byte;
         Col :Word;
         dum :Single;

PROCEDURE TwoRandomGauss(var N1, N2 : Single);
      { References:
        Box & Muller (1958): A note on the generation of random normal
                            deviates; Ann. of Math. Stat., Vol 29, p 610-611.
        Rubinstein (1981): Simulation and the Monte Carlo method; Wiley; p 86.
        Tjur, Tue (1990): Forelsningsnoter i Sandsynlighedsregning til brug
                                      ved kurset i Statistik 0, KUIMS; p 113. }
      Const
             TwoPi    = 6.28318530718;
             MinusTwo = -2.0;
      Var
             dum1, dum2
             : Single;
BEGIN { proc TwoRandomGauss }
        dum1:= sqrt( MinusTwo * Ln(Random) );
        dum2:= TwoPi * Random;
        N1:= dum1 * Cos(dum2);
        N2:= dum1 * Sin(dum2);
END;  { proc TwoRandomGauss }

BEGIN { proc CreateDeltaX }
  For Row:= 1 To DimBrownianMotion Do
    Begin
      For Col:= 1 To Trunc(T/2) Do
        TwoRandomGauss(DeltaX[Row]^[2*Col-1],DeltaX[Row]^[2*Col]);
      If Odd(T) Then
        TwoRandomGauss(DeltaX[Row]^[T],Dum);
    End;
END;  { proc CreateDeltaX }

PROCEDURE CreateX_;
  { 10.7.93
    This procedure cummulates the deltaX's
    and lay in the restricted variables }
VAR      RowTrendStat,RowX :Byte;
         Col :Word;
         dum:Real;
BEGIN { proc CreateX_ }
  For RowX:= 1 To DimUnResTrend Do
    Begin
      X_[RowX]^[1]:=0;
      For Col:= 2 To T Do
        Begin
          Dum:= DeltaX[RowX]^[Col-1] + Reg[RowX]^[Col-1];
          X_[RowX]^[Col]:= X_[RowX]^[Col-1] + Dum;
        End;
    End;

  For RowX:= (DimUnResTrend+1) To DimBrownianMotion Do
    Begin
      X_[RowX]^[1]:=0;
      For Col:= 2 To T Do
          X_[RowX]^[Col]:= X_[RowX]^[Col-1] + DeltaX[RowX]^[Col-1];
    End;

  For RowX:= (DimBrownianMotion+1) To DimX Do
    Begin
      RowTrendStat:= RowX - DimBrownianMotion;
      For Col:= 1 To T Do
        X_[RowX]^[Col]:= TrendStat[RowTrendStat]^[Col];
    End;

  If DimUnResTrend > 0 Then
    MultipelRegression(X_,Reg,X_,T,DimX,DimUnResTrend,RegReg);

END;  { proc CreateX_ }


PROCEDURE InitializeSimulation;
VAR      Row:Byte; Col:Word;
  { This procedure calculate the regressors and the restrictions, who are
    common for all simulation.
    Example:
                 DeltaX_t = Pi * X_t + mu + Epsilon_t,
                 alpha_ort * mu = 0
              Regressor  :  (1,1,1,1,1)'
              Restriction:  (0,0.25,0.5,0.75,1)'
  }

FUNCTION TestRank (var Cor: pxT_matrix; Row: Byte):Boolean;
VAR      Result:Boolean;
         Dummy:Real;
         Col:Word;
BEGIN { FUNCTION TestRank }
  Result:=False;
  Dummy:= 0.110669;                     { This value is choosed since it
                                          is strictly in the interval
                                          [0,1] and since it is my
                                          birthday }
  For Col:= 1 To (T-1) Do
  Dummy:= Dummy + Abs(Cor[Row]^[Col]);
  If Dummy > 0.5 Then Result:= True;

{  If Row>1 Then
        For i:= 1 To (CorrectRow-1) Do
          Begin
            Dummy:= 0;
            For Col:= 1 To T Do
              Dummy:= Dummy + Cor[CorrectRow]^[Col] * Cor[i]^[Col];
            If Abs(Dummy) > 0.0001 Then CorrectRow:= CorrectRow +1;
          End;
}
  TestRank:=Result;
END;  { FUNCTION TestRank }

PROCEDURE CreateCorrection(var Cor:pxT_matrix; var T:Word; var DimCor: Byte;
                           var WantCorrection: CorrectionType);
VAR     X1,i,Row,CorrectRow:Byte;
        X1T,Col:Word;
        T100,Dummy,Slope:Real;
BEGIN { subproc CreateCorrection }
  T100:=T/100;
  CorrectRow:=1;
  For Row:= 1 To DimCor Do
    Begin
      X1:= 0;
      For i:= 1 To 6 Do
        If WantCorrection[Row,i].X2 <> 0
        Then Begin
            Slope:= (WantCorrection[Row,i].Y2 - WantCorrection[Row,i].Y1)/
                    (WantCorrection[Row,i].X2 - X1)/T100;
            X1T:= Trunc(X1*T100);
            For Col:= X1T To Trunc(WantCorrection[Row,i].X2*T100) Do
              Begin
                Dummy:= WantCorrection[Row,i].Y1 + Slope*(Col-X1T);
                Cor[CorrectRow]^[Col]:= Dummy;
              End;
            X1:= WantCorrection[Row,i].X2;
          End;
      If TestRank(Cor,CorrectRow)=True Then CorrectRow:=CorrectRow+1;
    End;
  DimCor:= CorrectRow - 1;
END;  { subproc CreateCorrection }

BEGIN { proc InitializeSimulation }
  If FixedRandomStart Then
    RandSeed:=0
  Else
    Randomize;
  CreateCorrection(Reg,T,DimUnResTrend,WantReg);
  CreateCorrection(TrendStat,T,DimTrendStat,WantTrendStat);

  SquareToMatrix(Reg,RegReg,T,DimUnResTrend);
  InversePosDefMatrix(RegReg,DimUnResTrend);
END;  { proc InitializeSimulation }


PROCEDURE CalculateTest(var Trace:Single);
VAR      InvL11S10 :pxp_matrix;
BEGIN { proc CalculateTest }
  CalculateInvL11S10(X_,DeltaX,T,DimX,DimDeltaX,InvL11S10);
  TraceOfSquareToMatrix(InvL11S10,DimX,DimDeltaX,Trace);
END;  { proc CalculateTest }

PROCEDURE Simulation;
VAR      Trace :Single;
         Rep   :LongInt;
BEGIN { proc Simulation }
  GetMatrix(X_,DimX);
  GetMatrix(DeltaX,DimBrownianMotion);
  GetMatrix(Reg,DimUnResTrend);
  GetMatrix(TrendStat,DimTrendStat);
  InitializeSimulation;
  GoToXY(50,25); Write('Repetition no. ');
  Assign(EraseMe,'ERASE.ME');ReWrite(EraseMe);
    For Rep:= 1 To N Do
      Begin
        CreateDeltaX;
        CreateX_;
        CalculateTest(Trace);
        Write(EraseMe,Trace:12:6,'  ');
        GoToXY(66,25); Write(Rep:6);
      End;
    Writeln;
    Writeln(EraseMe);
  Close(EraseMe);
  DeleteMatrix(X_,DimX);
  DeleteMatrix(DeltaX,DimBrownianMotion);
  DeleteMatrix(Reg,DimUnResTrend);
  DeleteMatrix(TrendStat,DimTrendStat);
END;  { proc Simulation }


END.











